"""Specialization of `Dataset` for anomaly detection."""

import abc
import typing

import numpy

from .. import core


class AnomalyDataset(core.Dataset):
    """Specialization of `Dataset` for anomaly detection."""

    @property
    @abc.abstractmethod
    def labels(self) -> numpy.ndarray:
        """Labels of the instances in the data array."""
        pass


class AnomalyTabular(AnomalyDataset):
    """Example implementation of `AnomalyDataset` for tabular datasets."""

    def __init__(self, data: numpy.ndarray, scores: numpy.ndarray, name: str) -> None:
        """Initialize an `AnomalyTabular` dataset."""
        self.__name = name
        self.__data = data
        self.__scores = scores
        self.__indices = numpy.asarray(list(range(data.shape[0])), dtype=numpy.uint)

    @property
    def name(self) -> str:
        """Name of the dataset."""
        return self.__name

    @property
    def data(self) -> numpy.ndarray:
        """Data array."""
        return self.__data

    @property
    def labels(self) -> numpy.ndarray:
        """Labels of the rows in the data array."""
        return self.__scores

    @property
    def indices(self) -> numpy.ndarray:
        """Indices of the rows in the data array."""
        return self.__indices

    def __eq__(self, other: "AnomalyTabular") -> bool:  # type: ignore[override]
        """Check if two `AnomalyTabular` datasets have the same name."""
        return self.__name == other.__name

    @property
    def max_instance_size(self) -> int:
        """Maximum memory size of an instance in the dataset."""
        item_size = self.__data.itemsize
        num_items = numpy.prod(self.__data.shape[1:])
        return int(item_size * num_items)

    @property
    def approx_memory_size(self) -> int:
        """Approximate memory size of the dataset."""
        data_size = self.cardinality * self.max_instance_size
        scores_size = self.cardinality * self.__scores.itemsize
        return data_size + scores_size

    def __getitem__(
        self,
        item: typing.Union[int, typing.Iterable[int]],
    ) -> numpy.ndarray:
        """Get a row or multiple rows from the data array."""
        return self.__data[item]

    def subset(self, indices: list[int], subset_name: str) -> "AnomalyTabular":
        """Create a subset of the dataset."""
        data = self.__data[indices, :]
        scores = self.__scores[indices]
        return AnomalyTabular(data, scores, subset_name)


__all__ = [
    "AnomalyDataset",
    "AnomalyTabular",
]



"""This module defines the `Space` class."""

import abc
import random
import typing

import numpy

from . import dataset
from . import metric


class Space(abc.ABC):
    """This class combines a `Dataset` and a `Metric` into a `MetricSpace`.

    We build `Cluster`s and `Graph`s over a `MetricSpace`. This class provides
    access to the underlying `Dataset` and `Metric`. Subclasses should
    implement the methods to compute distances between indexed instances using
    the underlying `Metric` and `Dataset`.

    For those cases where the distance function in `Metric` is expensive to
    compute, this class provides a cache which stores the distance values
    between pairs of instances by their indices. If you want to use the cache,
    set the `use_cache` parameter to True when instantiating a metric space.
    """

    def __init__(self, use_cache: bool) -> None:
        """Initialize the `MetricSpace`."""
        self.__use_cache = use_cache
        self.__cache: dict[tuple[int, int], float] = {}

    @property
    def name(self) -> str:
        """The name of the `MetricSpace`."""
        return f"{self.data.name}__{self.distance_metric.name}"

    @property
    def uses_cache(self) -> bool:
        """Whether the object distance values."""
        return self.__use_cache

    @property
    @abc.abstractmethod
    def data(self) -> dataset.Dataset:
        """The `Dataset` used to compute distances between instances."""
        pass

    @property
    @abc.abstractmethod
    def distance_metric(self) -> metric.Metric:
        """The `Metric` used to compute distances between instances."""
        pass

    @abc.abstractmethod
    def are_instances_equal(self, left: int, right: int) -> bool:
        """Given two indices, returns whether the corresponding instances are equal.

        Usually, this will rely on using a `Metric`. Two equal instances
        should have a distance of zero between them. The default implementation
        relies on this.
        """
        return self.distance_one_to_one(left, right) == 0.0

    @abc.abstractmethod
    def subspace(self, indices: list[int], subset_data_name: str) -> "Space":
        """See the `Dataset.subset`."""
        pass

    @abc.abstractmethod
    def distance_one_to_one(self, left: int, right: int) -> float:
        """See the `Dataset.one_to_one`. The default implementation uses the cache."""
        if not self.is_in_cache(left, right):
            d = self.distance_metric.one_to_one(self.data[left], self.data[right])
            self.add_to_cache(left, right, d)
        return self.get_from_cache(left, right)

    @abc.abstractmethod
    def distance_one_to_many(self, left: int, right: list[int]) -> numpy.ndarray:
        """See the `Dataset.one_to_many`. The default implementation uses the cache."""
        distances = [self.distance_one_to_one(left, r) for r in right]
        return numpy.asarray(distances, dtype=numpy.float32)

    @abc.abstractmethod
    def distance_many_to_many(self, left: list[int], right: list[int]) -> numpy.ndarray:
        """See the `Dataset.many_to_many`. The default implementation uses the cache."""
        distances = [self.distance_one_to_many(i, right) for i in left]
        return numpy.stack(distances)

    @abc.abstractmethod
    def distance_pairwise(self, indices: list[int]) -> numpy.ndarray:
        """See the `Dataset.pairwise`. The default implementation uses the cache."""
        return self.distance_many_to_many(indices, indices)

    @staticmethod
    def __cache_key(i: int, j: int) -> tuple[int, int]:
        """This works because of the `symmetry` property of a `Metric`."""
        return (i, j) if i < j else (j, i)

    def is_in_cache(self, i: int, j: int) -> bool:
        """Checks whether the distance between `i` and `j` is in the cache."""
        return self.__cache_key(i, j) in self.__cache

    def get_from_cache(self, i: int, j: int) -> float:
        """Returns the distance between `i` and `j` from the cache.

        Raises a KeyError if the distance value is not in the cache.
        """
        return self.__cache[self.__cache_key(i, j)]

    def add_to_cache(self, i: int, j: int, distance: float) -> None:
        """Adds the given `distance` to the cache."""
        self.__cache[self.__cache_key(i, j)] = distance

    def remove_from_cache(self, i: int, j: int) -> float:
        """Removes the distance between `i` and `j` from the cache."""
        return self.__cache.pop(  # type: ignore[call-overload]
            self.__cache_key(i, j),
            default=0.0,
        )

    def clear_cache(self) -> int:
        """Empty the cache and return the number of items that were in the cache."""
        num_items = len(self.__cache)
        self.__cache.clear()
        return num_items

    def choose_unique(
        self,
        n: int,
        indices: typing.Optional[list[int]] = None,
    ) -> list[int]:
        """Randomly chooses `n` unique instances from the dataset.

        Args:
            n: The number of unique instances to choose.
            indices: Optional. The list of indices from which to choose.

        Returns:
            A randomly selected list of indices of `n` unique instances.
        """
        indices = indices or list(range(self.data.cardinality))

        if not (0 < n <= len(indices)):
            msg = (
                f"`n` must be a positive integer no larger than the length of "
                f"`indices` ({len(indices)}). Got {n} instead."
            )
            raise ValueError(msg)

        randomized_indices = indices.copy()
        random.shuffle(randomized_indices)

        if n == len(randomized_indices):
            return randomized_indices

        chosen: list[int] = []
        for i in randomized_indices:
            for o in chosen:
                if self.are_instances_equal(i, o):
                    break
            else:
                chosen.append(i)
                if len(chosen) == n:
                    break

        return chosen


class TabularSpace(Space):
    """A `Space` that uses a `TabularDataset` and a `Metric`."""

    def __init__(
        self,
        data: typing.Union[dataset.TabularDataset, dataset.TabularMMap],
        distance_metric: metric.Metric,
        use_cache: bool,
    ) -> None:
        """Initialize the `MetricSpace`."""
        self.__data = data
        self.__distance_metric = distance_metric
        super().__init__(use_cache)

    @property
    def data(
        self,
    ) -> typing.Union[dataset.TabularDataset, dataset.TabularMMap]:
        """Return the data."""
        return self.__data

    @property
    def distance_metric(self) -> metric.Metric:
        """Return the distance metric."""
        return self.__distance_metric

    def are_instances_equal(self, left: int, right: int) -> bool:
        """Return whether the instances at `left` and `right` are identical."""
        return self.distance_one_to_one(left, right) == 0.0

    def distance_one_to_one(self, left: int, right: int) -> float:
        """Compute the distance between `left` and `right`."""
        if self.uses_cache:
            return super().distance_one_to_one(left, right)
        return self.distance_metric.one_to_one(self.data[left], self.data[right])

    def distance_one_to_many(
        self,
        left: int,
        right: list[int],
    ) -> numpy.ndarray:
        """Compute the distances between `left` and each instance in `right`."""
        if self.uses_cache:
            return super().distance_one_to_many(left, right)
        return self.distance_metric.one_to_many(self.data[left], self.data[right])

    def distance_many_to_many(
        self,
        left: list[int],
        right: list[int],
    ) -> numpy.ndarray:
        """Compute the distances between instances in `left` and `right`."""
        if self.uses_cache:
            return super().distance_many_to_many(left, right)
        return self.distance_metric.many_to_many(self.data[left], self.data[right])

    def distance_pairwise(self, indices: list[int]) -> numpy.ndarray:
        """Compute the distances between all pairs of instances in `indices`."""
        return self.distance_metric.pairwise(self.data[indices])

    def subspace(
        self,
        indices: list[int],
        subset_data_name: str,
    ) -> "TabularSpace":
        """Return a subspace of this space."""
        return TabularSpace(
            self.data.subset(indices, subset_data_name),
            self.distance_metric,
            self.uses_cache,
        )


__all__ = [
    "Space",
    "TabularSpace",
]



"""Specialization of `Space` for anomaly detection."""

import abc

import numpy

from .. import core
from . import anomaly_dataset


class AnomalySpace(core.Space):
    """Specialization of `Space` for anomaly detection."""

    @property
    @abc.abstractmethod
    def data(self) -> anomaly_dataset.AnomalyDataset:
        """See the underlying `AnomalyDataset`."""
        pass

    @abc.abstractmethod
    def subspace(self, indices: list[int], subset_data_name: str) -> "AnomalySpace":
        """Return a subspace of the space."""
        pass


class AnomalyTabularSpace(AnomalySpace):
    """Example implementation of `AnomalySpace` for tabular datasets."""

    def __init__(
        self,
        data: anomaly_dataset.AnomalyDataset,
        distance_metric: core.Metric,
        use_cache: bool,
    ) -> None:
        """Initialize an `AnomalyTabularSpace`."""
        self.__data = data
        self.__distance_metric = distance_metric
        super().__init__(use_cache)

    @property
    def data(self) -> anomaly_dataset.AnomalyDataset:
        """See the underlying `AnomalyDataset`."""
        return self.__data

    def subspace(
        self,
        indices: list[int],
        subset_data_name: str,
    ) -> "AnomalyTabularSpace":
        """Return a subspace of the space."""
        return AnomalyTabularSpace(
            self.data.subset(indices, subset_data_name),  # type: ignore[arg-type]
            self.distance_metric,
            self.uses_cache,
        )

    @property
    def distance_metric(self) -> core.Metric:
        """Distance metric used to compare instances in the space."""
        return self.__distance_metric

    def are_instances_equal(self, left: int, right: int) -> bool:
        """Check if two instances are identical."""
        return self.distance_one_to_one(left, right) == 0.0

    def distance_one_to_one(self, left: int, right: int) -> float:
        """Compute the distance between two instances."""
        if self.uses_cache:
            return super().distance_one_to_one(left, right)
        return self.distance_metric.one_to_one(self.data[left], self.data[right])

    def distance_one_to_many(self, left: int, right: list[int]) -> numpy.ndarray:
        """Compute the distance between one instance and many instances."""
        if self.uses_cache:
            return super().distance_one_to_many(left, right)
        return self.distance_metric.one_to_many(self.data[left], self.data[right])

    def distance_many_to_many(self, left: list[int], right: list[int]) -> numpy.ndarray:
        """Compute the distance between many instances and many instances."""
        if self.uses_cache:
            return super().distance_many_to_many(left, right)
        return self.distance_metric.many_to_many(self.data[left], self.data[right])

    def distance_pairwise(self, indices: list[int]) -> numpy.ndarray:
        """Compute the distance between all pairs of instances."""
        return self.distance_metric.pairwise(self.data[indices])


__all__ = [
    "AnomalySpace",
    "AnomalyTabularSpace",
]



"""This module contains the `Cluster` class."""

import math
import typing

import numpy

from ..utils import constants
from ..utils import helpers
from . import cluster_criteria
from . import space

logger = helpers.make_logger(__name__)

Ratios = tuple[float, float, float, float, float, float]
RATIO_NAMES = ["cardinality", "radius", "lfd"]
RATIO_NAMES.extend([f"{name}_ema" for name in RATIO_NAMES])


# TODO: Figure out a theoretically sound way to have this as part of the Dataset class.
SUB_SAMPLE_LIMIT = 100


class Cluster:
    """A `Cluster` represents a collection of `"nearby"` instances in a metric space.

    A `Cluster` is a node in a binary tree we call the `Cluster Tree`.

    Ideally, a user will instantiate a `Cluster` using the `new_root` method and
    chain calls to the `build` and `partition` methods to create the full tree.
    We also provide an `iterative_partition` method to avoid the inefficiencies
    of recursion in Python.

    The name of a `Cluster` is uniquely determined by its location in the tree.
    Names are generated as they would be for a huffman tree. The root cluster
    has the name `"1"`. A left-child appends a `"0"` and a right-child appends
    a `"1"` to the name of the parent.
    """

    __slots__ = [
        "__metric_space",
        "__indices",
        "__name",
        "__parent",
        "__arg_samples",
        "__arg_center",
        "__arg_radius",
        "__radius",
        "__lfd",
        "__ratios",
        "__children",
        "__ancestry",
        "__subtree",
        "__candidate_neighbors",
    ]

    @staticmethod
    def new_root(metric_space: space.Space) -> "Cluster":
        """Creates a new root cluster for the given metric space."""
        return Cluster(
            metric_space,
            indices=list(range(metric_space.data.cardinality)),
            name="1",
            parent=None,
        )

    def __init__(
        self,
        metric_space: space.Space,
        *,
        indices: list[int],
        name: str,
        parent: typing.Optional["Cluster"],
    ) -> None:
        """Initializes a new Cluster."""
        if len(indices) == 0:
            msg = "Cannot instantiate a cluster with an empty list of `indices`."
            raise ValueError(
                msg,
            )

        self.__metric_space = metric_space
        self.__indices = indices
        self.__name = name
        self.__parent: typing.Optional["Cluster"] = parent

        self.__arg_samples: typing.Union[list[int], constants.Unset] = constants.UNSET
        self.__arg_center: typing.Union[int, constants.Unset] = constants.UNSET
        self.__arg_radius: typing.Union[int, constants.Unset] = constants.UNSET
        self.__radius: typing.Union[float, constants.Unset] = constants.UNSET
        self.__lfd: typing.Union[float, constants.Unset] = constants.UNSET
        self.__ratios: typing.Union[Ratios, constants.Unset] = constants.UNSET
        self.__children: typing.Optional[tuple["Cluster", "Cluster"]] = None
        self.__ancestry: typing.Union[
            list["Cluster"],
            constants.Unset,
        ] = constants.UNSET
        self.__subtree: typing.Union[
            list[list["Cluster"]],
            constants.Unset,
        ] = constants.UNSET
        self.__candidate_neighbors: typing.Union[
            dict["Cluster", float],
            constants.Unset,
        ] = constants.UNSET

    def __eq__(self, other: "Cluster") -> bool:  # type: ignore[override]
        """Clusters are equal if they have the same name."""
        return self.__name == other.__name

    def __lt__(self, other: "Cluster") -> bool:
        """Clusters are ordered by depth and then by name."""
        if self.depth == other.depth:
            return self.__name < other.__name
        return self.depth < other.depth

    def __str__(self) -> str:
        """Returns a string representation of the cluster."""
        return self.__name

    def __repr__(self) -> str:
        """Returns a string representation of the cluster."""
        return f"{self.metric_space} :: Cluster {self.__name}"

    def __hash__(self) -> int:
        """Hash the name of the cluster."""
        return hash(repr(self.__name))

    @property
    def metric_space(self) -> space.Space:
        """The metric space in which the cluster exists."""
        return self.__metric_space

    @property
    def indices(self) -> list[int]:
        """The indices of instances in the cluster."""
        return self.__indices

    @property
    def name(self) -> str:
        """The name (i.e. location) of the cluster in the tree."""
        return self.__name

    @property
    def depth(self) -> int:
        """The depth of the cluster in the tree."""
        return len(self.__name) - 1

    @property
    def cardinality(self) -> int:
        """The number of instances in the cluster."""
        return len(self.__indices)

    @property
    def arg_samples(self) -> list[int]:
        """The indices of sub-sampled instances used to find the center."""
        if self.__arg_samples is constants.UNSET:
            msg = "Please call `build` on the cluster before using this property."
            raise ValueError(
                msg,
            )
        return self.__arg_samples  # type: ignore[return-value]

    @property
    def arg_center(self) -> int:
        """Index of the cluster center."""
        if self.__arg_center is constants.UNSET:
            msg = "Please call `build` on the cluster before using this property."
            raise ValueError(
                msg,
            )
        return self.__arg_center  # type: ignore[return-value]

    @property
    def center(self) -> typing.Any:  # noqa: ANN401
        """The (approximate) geometric median of the cluster."""
        if self.__arg_center is constants.UNSET:
            msg = "Please call `build` on the cluster before using this property."
            raise ValueError(
                msg,
            )
        return self.__metric_space.data[self.__arg_center]  # type: ignore[index]

    @property
    def arg_radius(self) -> int:
        """Index of the farthest instance from the center."""
        if self.__arg_radius is constants.UNSET:
            msg = "Please call `build` on the cluster before using this property."
            raise ValueError(
                msg,
            )
        return self.__arg_radius  # type: ignore[return-value]

    @property
    def radius(self) -> float:
        """Distance form the center to the farthest instance in the cluster."""
        if self.__radius is constants.UNSET:
            msg = "Please call `build` on the cluster before using this property."
            raise ValueError(
                msg,
            )
        return self.__radius  # type: ignore[return-value]

    @property
    def is_singleton(self) -> bool:
        """Returns `True` if the cluster has zero radius.

        i.e. if the cluster has only one instance or all instances are identical.
        """
        if self.__radius is constants.UNSET:
            msg = "Please call `build` on the cluster before using this property."
            raise ValueError(
                msg,
            )
        return self.__radius == 0

    @property
    def lfd(self) -> float:
        """The local fractal dimension of the cluster."""
        if self.__lfd is constants.UNSET:
            msg = "Please call `build` on the cluster before using this property."
            raise ValueError(
                msg,
            )
        return self.__lfd  # type: ignore[return-value]

    @property
    def parent(self) -> typing.Optional["Cluster"]:
        """The parent of this cluster in the tree."""
        return self.__parent

    @property
    def ratios(self) -> Ratios:
        """The six cluster ratios used for mata machine learning in CHAODA.

        To build a correct set of ratios, the user need to call the
        `normalize_ratios` method on the root of a tree.
        """
        if self.__ratios is constants.UNSET:
            msg = "Please call `build` on the cluster before using this property."
            raise ValueError(
                msg,
            )
        return self.__ratios  # type: ignore[return-value]

    @property
    def is_leaf(self) -> bool:
        """Returns `True` if this cluster is a leaf in the tree."""
        return self.__children is None

    @property
    def children(self) -> typing.Optional[tuple["Cluster", "Cluster"]]:
        """The two children of this cluster."""
        return self.__children

    @property
    def max_leaf_depth(self) -> int:
        """The depth of the deepest leaf in the `subtree` of this cluster."""
        return self.subtree[-1][0].depth

    @property
    def left_child(self) -> "Cluster":
        """The left child of this cluster."""
        if self.__children is None:
            msg = "This Cluster is a leaf and has no children."
            raise ValueError(msg)
        return self.children[0]  # type: ignore[index]

    @property
    def right_child(self) -> "Cluster":
        """The right child of this cluster."""
        if self.__children is None:
            msg = "This Cluster is a leaf and has no children."
            raise ValueError(msg)
        return self.children[1]  # type: ignore[index]

    @property
    def subtree(self) -> list[list["Cluster"]]:
        """Returns a list of lists of all clusters in the subtree of this cluster.

        Each inner list contains clusters at the same depth.
        The subtree includes the cluster itself.
        """
        if self.__subtree is constants.UNSET:
            if self.is_leaf:
                self.__subtree = [[self]]
            else:
                left_subtree = self.left_child.subtree
                right_subtree = self.right_child.subtree

                if len(left_subtree) < len(right_subtree):
                    left_subtree.extend(
                        [] for _ in range(len(right_subtree) - len(left_subtree))
                    )

                if len(right_subtree) < len(left_subtree):
                    right_subtree.extend(
                        [] for _ in range(len(left_subtree) - len(right_subtree))
                    )

                self.__subtree = [[self]] + [
                    (left_layer + right_layer)
                    for left_layer, right_layer in zip(left_subtree, right_subtree)
                ]

        self.__subtree = [
            layer
            for layer in self.__subtree  # type: ignore[union-attr]
            if len(layer) > 0
        ]

        return self.__subtree

    @property
    def num_descendants(self) -> int:
        """Returns the number of Clusters in the subtree of this Cluster."""
        return sum(len(_c) for _c in self.subtree)

    @property
    def ancestry(self) -> list["Cluster"]:
        """Returns the path through the tree from the root to this Cluster."""
        if self.__ancestry is constants.UNSET:
            if self.__parent is None:
                self.__ancestry = []
            else:
                self.__ancestry = [*self.__parent.ancestry, self.__parent]
        return self.__ancestry  # type: ignore[return-value]

    @property
    def candidate_neighbors(self) -> dict["Cluster", float]:
        """Find clusters in the tree that could be neighbors.

        Or whose descendants could be neighbors, of this cluster or of its
        descendants if any subset of them were selected to build a `Graph`.

        This property helps optimize the process of finding edges when building
        `Graph`s.
        """
        if self.__candidate_neighbors is constants.UNSET:
            if self.__parent is None:
                self.__candidate_neighbors = {self: 0.0}
            else:
                candidates = self.__parent.candidate_neighbors
                radius = max(self.radius, candidates[self.__parent])

                non_leaf_candidates = [c for c in candidates if not c.is_leaf]
                if len(non_leaf_candidates) > 0:
                    children = [
                        child
                        for c in non_leaf_candidates
                        for child in c.children  # type: ignore[union-attr]
                    ]
                    arg_centers = [c.arg_center for c in children]
                    distances = list(
                        map(
                            float,
                            self.__metric_space.distance_one_to_many(
                                self.arg_center,
                                arg_centers,
                            ),
                        ),
                    )
                    candidates.update(
                        {
                            c: d
                            for c, d in zip(children, distances)
                            if d <= radius + c.radius * (1 + math.sqrt(2))
                        },
                    )

                candidates[self] = radius
                self.__candidate_neighbors = candidates

        return self.__candidate_neighbors  # type: ignore[return-value]

    def is_ancestor_of(self, other: "Cluster") -> bool:
        """Returns `True` if this cluster is an ancestor of `other`."""
        return (self.depth < other.depth) and (
            self.__name == other.__name[: len(self.__name)]
        )

    def is_descendent_of(self, other: "Cluster") -> bool:
        """Returns `True` if this cluster is a descendent of `other`."""
        return other.is_ancestor_of(self)

    def build(self) -> "Cluster":
        """Calculates and sets important properties of the Cluster.

        These include:
            - arg_samples
            - arg_center
            - center
            - arg_radius
            - radius
            - lfd (local fractal dimension)
            - ratios (before normalization).

        Returns:
            The modified `Cluster` with after calculating essential properties.
        """
        logger.debug(f"Building cluster {self.__name} ...")

        if self.cardinality < SUB_SAMPLE_LIMIT:
            self.__arg_samples = self.__indices.copy()
        else:
            n = int(math.sqrt(self.cardinality))
            self.__arg_samples = self.__metric_space.choose_unique(n, self.__indices)

        sample_distances = numpy.sum(
            self.__metric_space.distance_pairwise(self.__arg_samples),
            axis=0,
        )
        self.__arg_center = self.__arg_samples[numpy.argmin(sample_distances)]

        center_distances = self.__metric_space.distance_one_to_many(
            self.__arg_center,  # type: ignore[arg-type]
            self.__indices,
        )
        farthest = numpy.argmax(center_distances)
        self.__arg_radius = self.__indices[farthest]
        self.__radius = float(center_distances[farthest])

        if self.is_singleton:
            self.__lfd = 1
        else:
            half_count = sum(d <= (self.__radius / 2) for d in center_distances)
            if half_count == 0:
                msg = "half_count for singleton cluster was zero."
                raise ValueError(msg)
            if half_count == 1:
                self.__lfd = self.cardinality
            else:
                self.__lfd = self.cardinality / (math.log2(half_count))

        # Calculate cluster ratios
        if self.parent is None:
            self.__ratios = (1.0, 1.0, 1.0, 1.0, 1.0, 1.0)
        else:
            c = self.cardinality / self.parent.cardinality
            r = self.__radius / self.parent.radius
            l = self.__lfd / self.parent.lfd  # noqa: E741

            _, _, _, pc_, pr_, pl_ = self.parent.ratios
            c_ = helpers.next_ema(c, pc_)
            r_ = helpers.next_ema(r, pr_)
            l_ = helpers.next_ema(l, pl_)

            self.__ratios = (c, r, l, c_, r_, l_)

        return self

    def distance_to_other(self, other: "Cluster") -> float:
        """Compute the distance between the centers of the two Clusters."""
        return self.distance_to_indexed_instance(
            other.__arg_center,  # type: ignore[arg-type]
        )

    def distance_to_indexed_instance(self, index: int) -> float:
        """Compute the distance from the center of the Cluster to the instance."""
        return self.__metric_space.distance_one_to_one(
            self.__arg_center,  # type: ignore[arg-type]
            index,
        )

    def distance_to_instance(self, instance: typing.Any) -> float:  # noqa: ANN401
        """Compute the distance from the center of the Cluster to the instance."""
        return self.__metric_space.distance_metric.one_to_one(self.center, instance)

    def __can_be_partitioned(
        self,
        criteria: typing.Sequence[cluster_criteria.ClusterCriterion],
    ) -> bool:
        """Returns `True` if the cluster can be partitioned."""
        if isinstance(self.__arg_samples, constants.Unset):
            msg = (
                "Please call `build` on this Cluster before calling "
                "`partition` or `iterative_partition`."
            )
            raise ValueError(
                msg,
            )

        # Cannot partition a cluster with zero radius.
        # Every criterion must evaluate to True.
        return (not self.is_singleton) and all(c(self) for c in criteria)

    def iterative_partition(
        self,
        criteria: typing.Sequence[cluster_criteria.ClusterCriterion],
    ) -> "Cluster":
        """Iteratively partitions the cluster to leaves.

        This iteratively builds the tree in a breadth-first manner, avoiding
        the inefficiencies of recursion in Python.
        """
        if not self.__can_be_partitioned(criteria):
            return self

        layers: list[list["Cluster"]] = [[self]]

        while True:
            partitionable = [c for c in layers[-1] if c.__can_be_partitioned(criteria)]
            if len(partitionable) == 0:
                break

            logger.info(
                f"Partitioning {len(partitionable)} clusters at depth "
                f"{partitionable[0].depth} ...",
            )

            [c.partition(criteria, recursive=False) for c in partitionable]
            layers.append(
                [
                    child
                    for c in partitionable
                    for child in c.children  # type: ignore[union-attr]
                ],
            )

        self.__subtree = layers

        return self

    def partition(
        self,
        criteria: typing.Sequence[cluster_criteria.ClusterCriterion],
        recursive: bool = True,
    ) -> "Cluster":
        """Tries to partition this cluster into two child clusters.

        If the cluster gets partitioned, then the `children` property will be
        assigned as a tuple of the left and right child clusters.

        `partition` starts by selecting two highly separated instances from the
        cluster. These are the left and right poles. The remaining instances
        are then split into two sets. Instances closer to the left pole are in
        the left split and instances closer to the right pole are in the right
        split. The left and right sets, along with their respective poles, are
        each used to instantiate the child clusters. Each child cluster is then
        built and, optionally, partitioned with the same `criteria`.

        Args:
            criteria: A list of rules for partitioning a cluster. Each criterion
             will be called with `self` as the only argument. If any criterion
             returns `False`, then this cluster will not be partitioned.
            recursive: Whether to recursively partition child clusters.

        Returns:
            The modified cluster.
        """
        if not self.__can_be_partitioned(criteria):
            return self

        # The farthest instance from the center is used as one of the poles.
        left_pole = self.arg_radius
        remaining_indices = [i for i in self.__indices if i != left_pole]
        left_distances = self.__metric_space.distance_one_to_many(
            left_pole,
            remaining_indices,
        )

        # The farthest instance from the left pole is used as the right pole.
        arg_right = int(numpy.argmax(left_distances))
        right_pole: int = remaining_indices[arg_right]

        if len(remaining_indices) > 1:
            remaining_indices = [i for i in remaining_indices if i != right_pole]

            # For every instance that is not a pole, compute its distance to each pole.
            left_distances = numpy.asarray(
                [d for i, d in enumerate(left_distances) if i != arg_right],
                dtype=left_distances.dtype,
            )
            right_distances = self.__metric_space.distance_one_to_many(
                right_pole,
                remaining_indices,
            )

            is_closer_to_left_pole = [
                ld <= rd for ld, rd in zip(left_distances, right_distances)
            ]

            # The remaining instances are split by proximity to either pole.
            left_indices = [left_pole] + [
                i for i, b in zip(remaining_indices, is_closer_to_left_pole) if b
            ]
            right_indices = [right_pole] + [
                i for i, b in zip(remaining_indices, is_closer_to_left_pole) if not b
            ]

        else:
            left_indices = [i for i in self.__indices if i != right_pole]
            right_indices = [right_pole]

        # The children will be biased so that the cardinality of the left child
        # is no smaller than that of the right child.
        left_indices, right_indices = (
            (left_indices, right_indices)
            if len(left_indices) > len(right_indices)
            else (right_indices, left_indices)
        )

        # Recursively build and partition the left and right children.
        left_child = Cluster(
            self.__metric_space,
            indices=left_indices,
            name=self.__name + "0",
            parent=self,
        ).build()

        right_child = Cluster(
            self.__metric_space,
            indices=right_indices,
            name=self.__name + "1",
            parent=self,
        ).build()

        if recursive:
            left_child = left_child.partition(criteria)
            right_child = right_child.partition(criteria)

        self.__children = (left_child, right_child)

        return self

    def normalize_ratios(
        self,
        mode: helpers.NormalizationMode = "gaussian",
    ) -> "Cluster":
        """Normalize the cluster ratios in the `subtree` of this cluster.

        This method is intended to only be called once and only on the root.
        The user is responsible for this contract.

        Args:
            mode: Normalization method to use. Must be one of:
                - 'linear',
                - 'gaussian', or
                - 'sigmoid'.

        Returns:
            The modified cluster after setting the normalized ratios.
        """
        clusters = [c for layer in self.subtree for c in layer]
        ratios_array = numpy.asarray([c.ratios for c in clusters], dtype=numpy.float32)
        if ratios_array.shape != (len(clusters), len(RATIO_NAMES)):
            msg = (
                f"Expected ratios array to have shape "
                f"{len(clusters), len(RATIO_NAMES)}. "
                f"Got {ratios_array.shape} instead."
            )
            raise ValueError(msg)

        ratios_array = helpers.normalize(ratios_array, mode)

        for i, c in enumerate(clusters):
            c.__ratios = tuple(  # type: ignore[assignment]
                map(float, ratios_array[i, :]),
            )

        return self

    def add_instance(self, index: int) -> list["Cluster"]:
        """Add an instance to the tree after having added it to the metric space.

        This method may only be called on a root cluster. The primary use
        of this method is when the full dataset does not fit in memory. In such
        a case, the user can create a tree on a subset of the full data.
        Instances in the complement of the subset can then be added to the tree.

        The instance is recursively added to the child whose center is closer.

        This method will almost certainly invalidate cluster properties such as
        `center` and `radius`. It is upon the user to rebuild clusters after
        adding instances.

        Args:
            index: of the instance in the metric space.

        Returns:
            The list of clusters, from root to leaf, to which the instance was
             added.
        """
        # TODO: Figure out how and where an instance will be added to the metric space.

        if self.depth != 0:
            msg = "Cannot add an instance to a non-root Cluster."
            raise ValueError(msg)

        return sorted(self.__add_instance(index))

    def __add_instance(self, index: int) -> list["Cluster"]:
        """Recursively add the indexed instance to the tree.

        Appends `index` to `__indices` after addition.
        """
        result = [self]

        if not self.is_leaf:
            distance_to_left = self.left_child.distance_to_indexed_instance(index)
            distance_to_right = self.right_child.distance_to_indexed_instance(index)

            if distance_to_left <= distance_to_right:
                result.extend(self.left_child.__add_instance(index))
            else:
                result.extend(self.right_child.__add_instance(index))

        self.__indices.append(index)

        return result


__all__ = [
    "Cluster",
]



"""Synthetic data generation tests and examples in CLAM."""

import numpy

__all__ = [
    "random",
    "bullseye",
    "line",
    "xor",
    "spiral_2d",
    "tori",
    "skewer",
]

Data = tuple[numpy.ndarray, list[int]]


def random(n: int = 100, dimensions: int = 10) -> Data:
    """Generate a 2D random dataset."""
    rng = numpy.random.default_rng()
    data = rng.normal(size=(n, dimensions))
    labels = [0 for _ in range(n)]
    return data, labels


def ring_data(n: int, radius: float, noise: float) -> numpy.ndarray:
    """Generate a 2D ring dataset."""
    rng = numpy.random.default_rng()
    theta = 2 * numpy.pi * rng.random(n)
    x = radius * numpy.cos(theta) + noise * rng.normal(size=n)
    y = radius * numpy.sin(theta) + noise * rng.normal(size=n)
    ring = numpy.stack([x, y], axis=1)
    return numpy.asarray(ring, dtype=numpy.float64)


def bullseye(n: int = 2_000, num_rings: int = 3, noise: float = 0.05) -> Data:
    """Generate a 2D bullseye dataset."""
    data = numpy.ndarray(shape=(0, 2))
    labels = []
    for i, r in enumerate(range(1, 2 * num_rings, 2)):
        ring: numpy.ndarray = ring_data(n=n * r, radius=r, noise=noise)
        labels.extend([i for _ in range(n * r)])
        data = numpy.concatenate([data, ring], axis=0)
    return numpy.asarray(data, dtype=numpy.float64), labels


def line(n: int = 5_000, m: float = 1, c: float = 0.0, noise: float = 0.05) -> Data:
    """Generate a 2D line dataset."""
    rng = numpy.random.default_rng()
    x = rng.random(n)
    y = m * x + c
    data = numpy.asarray((x, y)).T
    data = data + rng.random(data.shape) * noise
    labels = numpy.ones_like(x.T)
    return numpy.asarray(data, dtype=numpy.float64), list(labels)


def xor(n: int = 5_000) -> Data:
    """Generate a 2D XOR dataset."""
    rng = numpy.random.default_rng()
    data = rng.random(n)
    labels = [int((x > 0.5) != (y > 0.5)) for x, y, in data]
    return numpy.asarray(data, dtype=numpy.float64), labels


def spiral_2d(n: int = 5_000, noise: float = 0.1) -> Data:
    """Generate a 2D spiral dataset."""
    rng = numpy.random.default_rng()

    theta = numpy.sqrt(rng.random(n)) * 2 * numpy.pi

    r_a = 2 * theta + numpy.pi
    data_a = numpy.array([numpy.cos(theta) * r_a, numpy.sin(theta) * r_a]).T
    x_a = data_a + rng.normal(size=n) * noise

    r_b = -2 * theta - numpy.pi
    data_b = numpy.array([numpy.cos(theta) * r_b, numpy.sin(theta) * r_b]).T
    x_b = data_b + rng.normal(size=n) * noise

    data = numpy.concatenate([x_a, x_b]) / 5
    labels = list(numpy.concatenate([numpy.zeros(len(x_a)), numpy.ones(len(x_a))]))
    return numpy.asarray(data, dtype=numpy.float64), labels


def generate_torus(
    n: int,
    r_torus: float,
    noise: float,
) -> tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]:
    """Generate a 3D torus."""
    rng = numpy.random.default_rng()

    r_tube: float = r_torus / 5
    u, v = rng.random(n), rng.random(n)
    u, v = u * 2 * numpy.pi, v * 2 * numpy.pi
    x = (r_torus + r_tube * numpy.cos(v)) * numpy.cos(u) + (rng.normal(size=n) * noise)
    y = (r_torus + r_tube * numpy.cos(v)) * numpy.sin(u) + (rng.normal(size=n) * noise)
    z = r_tube * numpy.sin(v) + (rng.normal(size=n) * noise)
    return x, y, z


def tori(n: int = 10_000, noise: float = 0.015, r_torus: float = 1.0) -> Data:
    """Generate a 3D tori dataset."""
    x, y, z = generate_torus(n=n // 2, r_torus=r_torus, noise=noise)
    torus_1 = numpy.stack([x - r_torus, y, z], axis=1)
    labels = [0 for _ in x]

    x, y, z = generate_torus(n=n // 2, r_torus=r_torus, noise=noise)
    torus_2 = numpy.stack([x, z, y], axis=1)
    labels.extend([1 for _ in x])

    data = numpy.concatenate([torus_1, torus_2], axis=0)
    return numpy.asarray(data, dtype=numpy.float64), labels


def spiral_3d(
    n: int,
    radius: float,
    height: float,
    num_turns: int,
    noise: float,
) -> numpy.ndarray:
    """Generate a 3D spiral dataset."""
    rng = numpy.random.default_rng()

    theta = 2 * numpy.pi * num_turns * rng.random(n)
    x = radius * numpy.cos(theta) + noise * rng.normal(size=n)
    y = radius * numpy.sin(theta) + noise * rng.normal(size=n)
    z = height * theta / (2 * numpy.pi * num_turns) + noise * rng.normal(size=n)
    return numpy.stack([x, y, z], axis=1)


def line_3d(n: int, height: float, noise: float) -> numpy.ndarray:
    """Generate a 3D line dataset."""
    rng = numpy.random.default_rng()

    x = noise * rng.normal(size=n)
    y = noise * rng.normal(size=n)
    z = height * rng.random(n)
    return numpy.stack([x, y, z], axis=1)


def skewer(
    n: int = 10_000,
    radius: float = 3.0,
    height: float = 6.0,
    num_turns: int = 2,
    noise: float = 0.05,
) -> Data:
    """Generate a 3D skewer dataset."""
    spiral_points, line_points = 5 * n // 6, n // 6
    spiral_data = spiral_3d(spiral_points, radius, height, num_turns, noise)
    labels = [0 for _ in range(spiral_data.shape[0])]

    line_data = line_3d(line_points, height, noise)
    labels.extend([1 for _ in range(line_data.shape[0])])

    data = numpy.concatenate([spiral_data, line_data], axis=0)
    return numpy.asarray(data, dtype=numpy.float64), labels



"""CAKES: CLAM Augmented K-nearest neighbors Entropy-scaling Search."""

import math
import typing

from ..core import cluster
from ..core import cluster_criteria
from ..core import dataset
from ..core import metric
from ..core import space
from ..utils import constants
from ..utils import helpers

logger = helpers.make_logger(__name__)

ClusterHits = dict[cluster.Cluster, float]
IndexedHits = dict[int, float]


class CAKES:
    """CLAM Augmented K-nearest neighbors Entropy-scaling Search."""

    def __init__(self, metric_space: space.Space) -> None:
        """Initialize a CAKES object."""
        self.__metric_space = metric_space
        self.__root = cluster.Cluster.new_root(metric_space)

        self.__depth = self.__root.max_leaf_depth

    @classmethod
    def from_root(cls, root: cluster.Cluster) -> "CAKES":
        """Creates a CAKES object from a given root cluster."""
        cakes = super().__new__(cls)
        cakes.__metric_space = root.metric_space
        cakes.__root = root
        cakes.__depth = root.max_leaf_depth
        return cakes

    @property
    def depth(self) -> int:
        """Return the depth of the search tree."""
        return self.__depth

    @property
    def root(self) -> cluster.Cluster:
        """Return the root cluster."""
        return self.__root

    @property
    def metric_space(self) -> space.Space:
        """Return the metric space."""
        return self.__metric_space

    @property
    def data(self) -> dataset.Dataset:
        """Return the dataset."""
        return self.__metric_space.data

    @property
    def distance_metric(self) -> metric.Metric:
        """Return the distance metric."""
        return self.__metric_space.distance_metric

    def build(
        self,
        max_depth: typing.Optional[int] = None,
        additional_criteria: typing.Optional[
            list[cluster_criteria.ClusterCriterion]
        ] = None,
    ) -> "CAKES":
        """Builds the search tree upto singleton leaves, or an optional maximum depth.

        Args:
            max_depth: Optional. maximum depth of search tree.
            additional_criteria: Optional. Any additional criteria to use for
             building the Cluster Tree.

        Returns:
            the modified CAKES object.
        """
        logger.info(
            "Building search tree to " + "leaves"
            if max_depth is None
            else f"max_depth {max_depth} ...",
        )

        depth_criterion = (
            cluster_criteria.NotSingleton()
            if max_depth is None
            else cluster_criteria.MaxDepth(max_depth)
        )
        criteria = [depth_criterion] + (additional_criteria or [])
        self.__root = self.root.build().iterative_partition(criteria)

        self.__depth = self.__root.max_leaf_depth

        return self

    def rnn_search(
        self,
        query_instance: typing.Any,  # noqa: ANN401
        search_radius: float,
    ) -> IndexedHits:
        """Performs rho-nearest neighbors search around query with given radius.

        Args:
            query_instance: instance around which to search.
            search_radius: distance from query from within which to return
             all hits.

        Returns:
            dictionary of index-of-neighbor -> distance-to-neighbor.
        """
        candidate_clusters: list[cluster.Cluster] = self.tree_search(
            query_instance,
            search_radius,
        )

        if len(candidate_clusters) == 0:
            return {}

        return self.leaf_search(query_instance, search_radius, candidate_clusters)

    def knn_search(
        self,
        query_instance: typing.Any,  # noqa: ANN401
        k: int,
    ) -> IndexedHits:
        """Performs k-nearest neighbors search around query.

        Args:
            query_instance: instance around which to search.
            k: number of closest neighbors to look for.

        Returns:
            dictionary of index-of-neighbor -> distance-to-neighbor.
        """
        if k < 1:
            msg = f"k must be a positive integer. Got {k} instead."
            raise ValueError(msg)

        search_radius = (k / self.__root.cardinality) * self.__root.radius
        if search_radius <= 0.0:
            msg = f"Expected positive search_radius. Got {search_radius:.2e} instead."
            raise ValueError(msg)

        hits = self.rnn_search(query_instance, search_radius)
        while len(hits) == 0:  # make sure to have non-zero hits
            search_radius *= 2.0
            hits = self.rnn_search(query_instance, search_radius)

        while len(hits) < k:  # make sure to have more at least k hits
            # calculate lfd in ball around query
            if len(hits) == 1:
                lfd = 1.0
            else:
                half_count = len(
                    [
                        point
                        for point, distance in hits.items()
                        if distance <= search_radius / 2
                    ],
                )
                lfd = 1.0 if half_count == 0 else math.log2(len(hits) / half_count)

            # increase radius as guided by lfd
            if lfd < 1e-3:
                factor = 2
            else:
                factor = (k / len(hits)) ** (1.0 / (lfd + constants.EPSILON))
            if factor <= 1.0:
                msg = f"expected factor to be greater than 1. Got {factor:.2e} instead."
                raise ValueError(msg)

            search_radius *= factor

            # rerun rnn search
            hits = self.rnn_search(query_instance, search_radius)

        # sort hits in non-decreasing order of distance to query
        results = sorted([(distance, point) for point, distance in hits.items()])
        return {point: distance for distance, point in results[:k]}

    def tree_search(
        self,
        query_instance: typing.Any,  # noqa: ANN401
        search_radius: float,
    ) -> list[cluster.Cluster]:
        """Performs tree-search for the query, starting at the root.

        Consider the sphere centered at `query_instance` with the given
        `search_radius`. This method performs a breadth-first search over the
        tree and, at each iteration, discards the clusters whose volumes do not
        overlap with this sphere. Once the search terminates, the remaining
        clusters are returned.

        Args:
            query_instance: around which to search.
            search_radius: within which to look.

        Returns:
            dictionary of cluster -> distance-from-center-to-query
        """
        return self.tree_search_history(query_instance, search_radius)[1]

    def tree_search_history(
        self,
        query_instance: typing.Any,  # noqa: ANN401
        search_radius: float,
    ) -> tuple[ClusterHits, list[cluster.Cluster]]:
        """Performs tree-search for the query, starting at the root.

        Also returns the full history.
        """
        history: ClusterHits = {}
        hits: list[cluster.Cluster] = []
        candidates = [self.root]

        while len(candidates) > 0:
            logger.debug(f"Searching tree at depth {candidates[0].depth} ...")
            centers = [self.data[c.arg_center] for c in candidates]
            distances = list(
                map(float, self.distance_metric.one_to_many(query_instance, centers)),
            )
            close_enough: ClusterHits = {
                c: d
                for c, d in zip(candidates, distances)
                if d <= (c.radius + search_radius)
            }
            history.update(close_enough)
            terminal = {
                c
                for c, d in close_enough.items()
                if c.is_leaf or (c.radius + d) <= search_radius
            }
            hits.extend(terminal)
            candidates = [
                child
                for c in close_enough
                if c not in terminal
                for child in c.children  # type: ignore[union-attr]
            ]

        return history, hits

    def leaf_search(
        self,
        query_instance: typing.Any,  # noqa: ANN401
        search_radius: float,
        candidate_clusters: list[cluster.Cluster],
    ) -> IndexedHits:
        """Performs leaf search for query on given clusters.

        Args:
            query_instance: around which to look.
            search_radius: within which to look.
            candidate_clusters: candidate clusters which need to be exhaustively
             searched.

        Returns:
            dictionary of index-of-hit-instance -> distance-to-hit.
        """
        # get indices of all points in candidate clusters
        indices = [i for c in candidate_clusters for i in c.indices]
        logger.debug(f"Performing leaf-search over {len(indices)} instances.")

        # get distances from query to candidate points
        instances = [self.data[i] for i in indices]
        distances = list(
            map(float, self.distance_metric.one_to_many(query_instance, instances)),
        )

        # Filter hits with points within radius of query
        return {i: d for i, d in zip(indices, distances) if d <= search_radius}


__all__ = [
    "CAKES",
]



"""Helper functions for the package."""

import logging
import typing

import numpy
from scipy.special import erf

from . import constants

NormalizationMode = typing.Literal[
    "linear",
    "gaussian",
    "sigmoid",
]


def make_logger(name: str) -> logging.Logger:
    """Create a logger with the given name."""
    logger = logging.getLogger(name)
    logger.setLevel(constants.LOG_LEVEL)
    return logger


def next_ema(ratio: float, ema: float) -> float:
    """Computes the next exponential moving average."""
    return constants.EMA_ALPHA * ratio + (1 - constants.EMA_ALPHA) * ema


def normalize(values: numpy.ndarray, mode: NormalizationMode) -> numpy.ndarray:
    """Normalizes each column in values into a [0, 1] range.

    Args:
        values: A 1-d or 2-d array of values to normalize.
        mode: Normalization mode to use. Must be one of:
         - 'linear',
         - 'gaussian', or
         - 'sigmoid'.

    Returns:
        array of normalized values.
    """
    squeeze = False
    if len(values.shape) == 1:
        squeeze = True
        values = numpy.expand_dims(values, axis=1)

    if mode == "linear":
        min_v, max_v = numpy.min(values, axis=0), numpy.max(values, axis=0)
        values = (values - min_v) / (max_v - min_v + constants.EPSILON)
    else:
        means = numpy.mean(values, axis=0)
        sds = numpy.std(values, axis=0) + constants.EPSILON
        if mode == "gaussian":
            values = (1 + erf((values - means) / (sds * numpy.sqrt(2)))) / 2
        else:
            values = 1 / (1 + numpy.exp(-(values - means) / sds))

    values = values.clip(constants.EPSILON, 1)
    if squeeze:
        values = numpy.squeeze(values)
    return values



"""This module contains the (experimental) Classifier class."""

import operator
import typing

import numpy

from .. import core
from ..anomaly_detection import CHAODA
from ..utils import helpers

logger = helpers.make_logger(__name__)


class Classifier:
    """A CLAM Classifier.

    This is very experimental.
    """

    def __init__(
        self,
        labels: numpy.ndarray,
        metric_spaces: typing.Sequence[core.Space],
        **kwargs,  # noqa: ANN003
    ) -> None:
        """Creates and initializes a CLAM Classifier.

        Lower scores are better.

        Args:
            metric_spaces: See `CHAODA`.
            labels: 1d array of labels. dtype must be numpy.uint.
            kwargs: These are the same as the kwargs for `CHAODA`.
        """
        if labels.dtype != numpy.uint:
            msg = f"labels must have dtype {numpy.uint}. Got {labels.dtype} instead."
            raise ValueError(
                msg,
            )

        self.__metric_spaces = metric_spaces
        self.__labels = list(map(int, labels))
        self.__unique_labels = set(self.__labels)
        self.__kwargs = kwargs
        self.__bowls: dict[int, CHAODA] = {}

    @property
    def labels(self) -> list[int]:
        """Returns the labels in the dataset."""
        return self.__labels

    @property
    def unique_labels(self) -> set[int]:
        """Returns the unique labels in the dataset."""
        return self.__unique_labels

    def build(self) -> "Classifier":
        """Fits the Classifier to the data and returns the fitted object."""
        for label in self.__unique_labels:
            logger.info(f"Fitting CHAODA object for label {label} ...")

            indices = [i for i, _l in enumerate(self.__labels) if _l == label]
            metric_spaces = [
                s.subspace(indices, f"{s.data.name}__{label}")
                for s in self.__metric_spaces
            ]

            self.__bowls[label] = CHAODA(metric_spaces, **self.__kwargs).build()

        return self

    def rank_single(self, query: typing.Any) -> list[tuple[int, float]]:  # noqa: ANN401
        """Predicts the class rankings for a single query."""
        label_scores = []
        for label, bowl in self.__bowls.items():
            score = bowl.predict_single(query)
            label_scores.append((label, score))

        return label_scores

    def rank(self, queries: core.Dataset) -> list[list[tuple[int, float]]]:
        """Predicts the class rankings for a set of queries."""
        label_scores = []
        for i in range(queries.cardinality):
            logger.info(f"Predicting class for query {i} ...")
            label_scores.append(self.rank_single(queries[i]))
        return label_scores

    def predict_single(self, query: typing.Any) -> tuple[int, float]:  # noqa: ANN401
        """Predicts the label for a single query."""
        label_scores = self.rank_single(query)
        best_label, best_score = min(label_scores, key=operator.itemgetter(1))
        return best_label, best_score

    def predict(self, queries: core.Dataset) -> tuple[list[int], list[float]]:
        """Predicts the label for a set of queries."""
        label_scores = []
        for i in range(queries.cardinality):
            logger.info(f"Predicting class for query {i + 1}/{queries.cardinality} ...")
            label_scores.append(self.predict_single(queries[i]))

        [labels, scores] = list(zip(*label_scores))
        return labels, scores  # type: ignore[return-value]


__all__ = [
    "Classifier",
]



"""Graph criteria for selecting clusters from a tree to make graphs."""

import abc
import heapq
import typing

import numpy

from ..utils import helpers
from . import cluster

logger = helpers.make_logger(__name__)


class GraphCriterion(abc.ABC):
    """Criterion for selecting clusters from a tree to make graphs.

    A `GraphCriterion` can be called on the `root` cluster of a tree to
    select a set of clusters from the tree. These clusters can be used to
    create a `Graph`. Subclasses much implement the `select` method. The
    abstract `GraphCriterion` will verify that the set of selected clusters
    obey the graph invariant. See the documentation for `Graph` for details on
    the graph invariant.
    """

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Returns the name of the criterion."""
        pass

    @abc.abstractmethod
    def select(self, root: cluster.Cluster) -> set[cluster.Cluster]:
        """Selects a set of clusters from the tree rooted at `root`."""
        pass

    @staticmethod
    def assert_invariant(root: cluster.Cluster, selected: set[cluster.Cluster]) -> None:
        """Verifies that the set of selected clusters obey the graph invariant."""
        for c in selected:
            if any(c.is_ancestor_of(other) for other in selected):
                msg = "A cluster and its ancestor were both selected."
                raise ValueError(msg)

        indices = {i for c in selected for i in c.indices}

        if len(indices) != root.cardinality:
            msg = (
                f"There was a mis-match in the number of instances that were selected. "
                f"The selected clusters have {len(indices)} instance but the "
                f"root has {root.cardinality}."
            )
            raise ValueError(msg)

    def __call__(self, root: cluster.Cluster) -> set[cluster.Cluster]:
        """Selects a set of clusters from the tree rooted at `root`."""
        selected = self.select(root)
        self.assert_invariant(root, selected)
        return selected


class Layer(GraphCriterion):
    """Selects the layer at the specified depth, -1 means go to leaves."""

    def __init__(self, depth: int) -> None:
        """Initializes a `Layer` criterion."""
        if depth < -1:
            msg = f"expected a '-1' or a non-negative depth. got: {depth}"
            raise ValueError(msg)
        self.depth = depth

    @property
    def name(self) -> str:
        """Returns the name of the criterion."""
        return f"Layer_{self.depth}"

    def select(self, root: cluster.Cluster) -> set[cluster.Cluster]:
        """Selects the layer at the specified depth, -1 means go to leaves."""
        if self.depth == -1:
            return {c for layer in root.subtree for c in layer if c.is_leaf}

        selected = {
            c for layer in root.subtree[: self.depth] for c in layer if c.is_leaf
        }
        selected.update(root.subtree[self.depth])

        return selected


class PropertyThreshold(GraphCriterion):
    """Selects clusters when the given property crosses the given percentile."""

    def __init__(
        self,
        value: typing.Literal["cardinality", "radius", "lfd"],
        percentile: float,
        mode: typing.Literal["above", "below"],
    ) -> None:
        """Initializes a `PropertyThreshold` criterion.

        During a BFT of the tree, this keeps track of the given cluster
        property. A cluster is selected if any of the following is true:
            - it is a leaf, or
            - it qualifies by `mode`, or
            - both children qualify by `mode`.

        If `mode` is 'above', a cluster qualifies if the property goes above the
        `percentile`. If `mode` is 'below', a cluster qualifies if the property
        goes below the `percentile`.

        Args:
            value: The cluster property to keep track of. Must be one of
             'cardinality', 'radius', or 'lfd'.
            percentile: the percentile, in the (0, 100) range, to be crossed.
            mode: select a cluster when `value` is 'above' or 'below' the
             `percentile`.
        """
        if 0.0 < percentile < 100.0:
            self.percentile: float = percentile
        else:
            msg = (
                f"percentile must be in the (0, 100) range. "
                f"Got {percentile:.2f} instead."
            )
            raise ValueError(msg)

        self.value = value

        if mode == "above":
            self.qualifies = lambda c, v: getattr(c, self.value) > v
        elif mode == "below":
            self.qualifies = lambda c, v: getattr(c, self.value) < v
        else:
            msg = f"mode must be 'above' or 'below'. Got {mode} instead."
            raise ValueError(msg)

    @property
    def name(self) -> str:
        """Returns the name of the criterion."""
        return f"PropertyThreshold_{self.value}_{self.name}_{self.percentile:.2f}"

    def select(self, root: cluster.Cluster) -> set[cluster.Cluster]:
        """Selects clusters when the given property crosses the given threshold."""
        threshold = float(
            numpy.percentile(
                [
                    getattr(c, self.value)
                    for layer in root.subtree
                    for c in layer
                    if c.cardinality > 1
                ],
                self.percentile,
            ),
        )
        selected: set[cluster.Cluster] = set()
        frontier: set[cluster.Cluster] = {root}
        while frontier:
            c = frontier.pop()
            if (
                c.is_leaf  # select the leaves
                or self.qualifies(c, threshold)  # if cluster qualifies
                or (
                    self.qualifies(c.left_child, threshold)
                    and self.qualifies(c.right_child, threshold)
                )
            ):
                selected.add(c)
            else:  # add children to frontier
                frontier.update(c.children)  # type: ignore[arg-type]
        return selected


class MetaMLSelect(GraphCriterion):
    """Uses the scoring function from a trained meta-ml model to select clusters."""

    def __init__(
        self,
        scorer: typing.Callable[[numpy.ndarray], float],
        name: typing.Optional[str] = None,
        min_depth: int = 4,
    ) -> None:
        """Initializes a `MetaMLSelect` criterion.

        Args:
            scorer: A function that takes the ratios of a cluster and returns
            its predicted score. Higher scoring clusters are considered
            better than lower scoring clusters.
            name: The name of the criterion.
            min_depth: The minimum depth in the tree for a cluster to be
            selected.
        """
        if min_depth < 1:
            msg = "min-depth must be a positive integer."
            raise ValueError(msg)

        self.__name = scorer.__name__ if name is None else name
        self.scorer = lambda c: -scorer(numpy.asarray(c.ratios, dtype=numpy.float32))
        self.min_depth = min_depth

    @property
    def name(self) -> str:
        """Returns the name of the criterion."""
        return self.__name

    def select(self, root: cluster.Cluster) -> set[cluster.Cluster]:
        """Selects clusters using the scoring function."""
        tree = [c for layer in root.subtree for c in layer]

        candidate_clusters = [c for c in tree if c.depth >= self.min_depth]
        normalized_scores = list(
            map(
                float,
                helpers.normalize(
                    numpy.asarray(list(map(self.scorer, candidate_clusters))),
                    mode="gaussian",
                ),
            ),
        )

        heap = list(zip(normalized_scores, candidate_clusters))
        heapq.heapify(heap)

        selected = {c for c in tree if c.depth < self.min_depth and c.is_leaf}
        selected_indices = {i for c in selected for i in c.indices}

        while len(heap) > 0:
            _, c = heapq.heappop(heap)

            if len(selected_indices.intersection(set(c.indices))) > 0:
                continue
            selected.add(c)
            selected_indices.update(set(c.indices))

        return selected


__all__ = [
    "GraphCriterion",
    "Layer",
    "PropertyThreshold",
    "MetaMLSelect",
]



"""Graph anomaly detection algorithms."""

import abc
import typing

import numpy

from .. import core
from ..utils import constants
from ..utils import helpers

logger = helpers.make_logger(__name__)

ClusterScores = dict[core.Cluster, float]
InstanceScores = dict[int, float]

# TODO: Look at DSD (diffusion-state-distance) as a possible individual-method.
# TODO: Specialize `should_be_fast` for each algorithm based on complexity analysis
# and graph properties.


class GraphScorer(abc.ABC):
    """A base class for all anomaly detection algorithms."""

    def __call__(self, g: core.Graph) -> tuple[ClusterScores, numpy.ndarray]:
        """Use this scorer to generate normalized anomaly scores."""
        logger.info(
            f"Running method {self.name} on a "
            "graph with {g.vertex_cardinality} clusters.",
        )

        cluster_scores = self.score_graph(g)
        if self.normalize_on_clusters:
            cluster_scores = self.__normalize_cluster_scores(cluster_scores)

        instance_scores = self.inherit_scores(cluster_scores)
        if not self.normalize_on_clusters:
            instance_scores = self.__normalize_instance_scores(instance_scores)

        scores_array = self.ordered_scores(g, instance_scores)
        return cluster_scores, scores_array

    @abc.abstractmethod
    def __hash__(self) -> int:
        """A way to uniquely identify each `GraphScorer` object."""
        pass

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Full name of the algorithm."""
        pass

    @property
    @abc.abstractmethod
    def short_name(self) -> str:
        """Abbreviated name of the algorithm."""
        pass

    @property
    @abc.abstractmethod
    def normalize_on_clusters(self) -> bool:
        """Whether to normalize scores for clusters or for instances."""
        pass

    @property
    @abc.abstractmethod
    def normalization_mode(self) -> helpers.NormalizationMode:
        """What normalization method to use."""
        pass

    @abc.abstractmethod
    def score_graph(self, g: core.Graph) -> ClusterScores:
        """The main method of the algorithm.

        This should return a dictionary whose keys are `Clusters` in the `Graph`
        and whose values are the pre-normalization anomaly scores.
        """
        pass

    @abc.abstractmethod
    def should_be_fast(self, g: core.Graph) -> bool:
        """Whether this algorithm is expected to run in a reasonably short time.

        This estimate method should made quickly.
        """
        pass

    def __normalize_cluster_scores(
        self,
        scores: ClusterScores,
    ) -> ClusterScores:
        """Normalize scores to be in the [0, 1] range."""
        [keys, scores] = list(zip(*scores.items()))  # type: ignore[assignment]
        scores = helpers.normalize(
            numpy.asarray(scores, dtype=numpy.float32),
            mode=self.normalization_mode,
        )
        return dict(zip(keys, (float(x) for x in scores)))  # type: ignore[arg-type]

    def __normalize_instance_scores(
        self,
        scores: InstanceScores,
    ) -> InstanceScores:
        """Normalize scores to be in the [0, 1] range."""
        [keys, scores] = list(zip(*scores.items()))  # type: ignore[assignment]
        scores = helpers.normalize(
            numpy.asarray(scores, dtype=numpy.float32),
            mode=self.normalization_mode,
        )
        return dict(zip(keys, (float(x) for x in scores)))  # type: ignore[arg-type]

    @staticmethod
    def inherit_scores(scores: ClusterScores) -> InstanceScores:
        """Inherit cluster scores to instance scores."""
        return {i: s for c, s in scores.items() for i in c.indices}

    @staticmethod
    def ordered_scores(g: core.Graph, scores: InstanceScores) -> numpy.ndarray:
        """Return scores in the order of the graph's vertices."""
        [indices, scores_] = list(zip(*sorted(scores.items())))
        if set(indices) != {i for c in g.clusters for i in c.indices}:
            msg = "Graph invariant violated."
            raise ValueError(msg)
        return numpy.asarray(scores_, dtype=numpy.float32)


class ClusterCardinality(GraphScorer):
    """A method that scores clusters by their cardinality.

    The smaller the cluster, the more anomalous it is.
    """

    def __hash__(self) -> int:
        """A way to uniquely identify each `GraphScorer` object by its name."""
        return hash(self.name)

    @property
    def name(self) -> str:
        """Full name of the algorithm."""
        return "cluster_cardinality"

    @property
    def short_name(self) -> str:
        """Abbreviated name of the algorithm."""
        return "cc"

    @property
    def normalize_on_clusters(self) -> bool:
        """Whether to normalize scores for clusters."""
        return True

    @property
    def normalization_mode(self) -> helpers.NormalizationMode:
        """The normalization mode to use."""
        return "gaussian"

    def should_be_fast(self, _: core.Graph) -> bool:
        """This method is expected to be fast for all graphs."""
        return True

    def score_graph(self, g: core.Graph) -> ClusterScores:
        """The main method of the algorithm."""
        return {c: -c.cardinality for c in g.clusters}


class ComponentCardinality(GraphScorer):
    """A method that scores clusters by their component's cardinality.

    The smaller the component, the more anomalous its clusters are.
    """

    def __hash__(self) -> int:
        """A way to uniquely identify each `GraphScorer` object by its name."""
        return hash(self.name)

    @property
    def name(self) -> str:
        """Full name of the algorithm."""
        return "component_cardinality"

    @property
    def short_name(self) -> str:
        """Abbreviated name of the algorithm."""
        return "sc"

    @property
    def normalize_on_clusters(self) -> bool:
        """Whether to normalize scores for clusters."""
        return True

    @property
    def normalization_mode(self) -> helpers.NormalizationMode:
        """The normalization mode to use."""
        return "gaussian"

    def should_be_fast(self, _: core.Graph) -> bool:
        """This method is expected to be fast for all graphs."""
        return True

    def score_graph(self, g: core.Graph) -> ClusterScores:
        """The main method of the algorithm."""
        return {
            c: -component.vertex_cardinality
            for component in g.components
            for c in component.clusters
        }


class VertexDegree(GraphScorer):
    """A method that scores clusters by their vertex degree.

    The smaller the degree, the more anomalous the cluster is.
    """

    def __hash__(self) -> int:
        """A way to uniquely identify each `GraphScorer` object by its name."""
        return hash(self.name)

    @property
    def name(self) -> str:
        """Full name of the algorithm."""
        return "vertex_degree"

    @property
    def short_name(self) -> str:
        """Abbreviated name of the algorithm."""
        return "vd"

    @property
    def normalize_on_clusters(self) -> bool:
        """Whether to normalize scores for clusters."""
        return True

    @property
    def normalization_mode(self) -> helpers.NormalizationMode:
        """The normalization mode to use."""
        return "gaussian"

    def should_be_fast(self, _: core.Graph) -> bool:
        """This method is expected to be fast for all graphs."""
        return True

    def score_graph(self, g: core.Graph) -> ClusterScores:
        """The main method of the algorithm."""
        return {c: -g.vertex_degree(c) for c in g.clusters}


class ParentCardinality(GraphScorer):
    """A method that scores clusters by their parent's cardinality.

    The smaller the cluster relative to its parent, the more anomalous the cluster is.
    """

    def __init__(self, depth_weight: typing.Callable[[int], float]) -> None:
        """Initialize the scorer."""
        self.weight = depth_weight

    def __hash__(self) -> int:
        """A way to uniquely identify each `GraphScorer` object by its name."""
        return hash(self.name)

    @property
    def name(self) -> str:
        """Full name of the algorithm."""
        return "parent_cardinality"

    @property
    def short_name(self) -> str:
        """Abbreviated name of the algorithm."""
        return "pc"

    @property
    def normalize_on_clusters(self) -> bool:
        """Whether to normalize scores for clusters."""
        return True

    @property
    def normalization_mode(self) -> helpers.NormalizationMode:
        """The normalization mode to use."""
        return "gaussian"

    def should_be_fast(self, _: core.Graph) -> bool:
        """This method is expected to be fast for all graphs."""
        return True

    def score_graph(self, g: core.Graph) -> ClusterScores:
        """The main method of the algorithm."""
        scores = {c: 0.0 for c in g.clusters}

        for c in g.clusters:
            ancestry = c.ancestry
            for i in range(1, len(ancestry)):
                scores[c] += (
                    self.weight(i)
                    * ancestry[i - 1].cardinality
                    / ancestry[i].cardinality
                )

        return {c: scores[c] for c in g.clusters}


class GraphNeighborhood(GraphScorer):
    """A method that scores clusters by the number of clusters in their neighborhood.

    The smaller the neighborhood, the more anomalous the cluster is.
    """

    def __init__(self, eccentricity_fraction: float) -> None:
        """Initialize the scorer."""
        if not (0 < eccentricity_fraction <= 1):
            msg = (
                "eccentricity fraction must be in the (0, 1] range. "
                "Got {eccentricity_fraction:.2e} instead."
            )
            raise ValueError(
                msg,
            )

        self.eccentricity_fraction = eccentricity_fraction

    def __hash__(self) -> int:
        """A way to uniquely identify each `GraphScorer` object by its name."""
        return hash(self.name)

    @property
    def name(self) -> str:
        """Full name of the algorithm."""
        return "graph_neighborhood"

    @property
    def short_name(self) -> str:
        """Abbreviated name of the algorithm."""
        return "gn"

    @property
    def normalize_on_clusters(self) -> bool:
        """Whether to normalize scores for clusters."""
        return True

    @property
    def normalization_mode(self) -> helpers.NormalizationMode:
        """The normalization mode to use."""
        return "gaussian"

    def should_be_fast(self, g: core.Graph) -> bool:
        """This method is expected to be fast for small graphs."""
        return g.vertex_cardinality < 256

    def num_steps(self, g: core.Graph, c: core.Cluster) -> int:
        """The number of steps to take away from the cluster."""
        return int(g.eccentricity(c) * self.eccentricity_fraction) + 1

    def score_graph(self, g: core.Graph) -> ClusterScores:
        """The main method of the algorithm."""
        step_sizes: dict[core.Cluster, list[int]] = {
            c: g.frontier_sizes(c)[: self.num_steps(g, c)] for c in g.clusters
        }
        return {c: -float(sum(sizes)) for c, sizes in step_sizes.items()}


class StationaryProbabilities(GraphScorer):
    """A method that scores clusters by their stationary probabilities.

    The stationary probability of a cluster is the probability that a random walk
    starting from a vertex in the cluster will end up in the cluster after a large
    number of steps.

    The smaller the stationary probability, the more anomalous the cluster is.
    """

    def __init__(self, steps: int) -> None:
        """Initialize the scorer."""
        self.steps = steps

    def __hash__(self) -> int:
        """A way to uniquely identify each `GraphScorer` object by its name."""
        return hash(self.name)

    @property
    def name(self) -> str:
        """Full name of the algorithm."""
        return "stationary_probabilities"

    @property
    def short_name(self) -> str:
        """Abbreviated name of the algorithm."""
        return "sp"

    @property
    def normalize_on_clusters(self) -> bool:
        """Whether to normalize scores for clusters."""
        return True

    @property
    def normalization_mode(self) -> helpers.NormalizationMode:
        """The normalization mode to use."""
        return "gaussian"

    def should_be_fast(self, _: core.Graph) -> bool:
        """This method is expected to be fast for small graphs."""
        return True

    def score_graph(self, g: core.Graph) -> ClusterScores:
        """The main method of the algorithm."""
        scores = {}

        for component in g.components:
            if component.vertex_cardinality > 1:
                clusters, matrix = component.distance_matrix

                # TODO: Figure out how to go until convergence.
                sums = numpy.sum(matrix, axis=1)[:, None] + constants.EPSILON
                matrix = matrix / sums
                for _ in range(self.steps):
                    matrix = numpy.linalg.matrix_power(matrix, 2)

                scores.update(
                    dict(zip(clusters, numpy.sum(matrix, axis=0))),
                )

            else:  # Component with only one Cluster means the Cluster is anomalous.
                scores.update({c: 0 for c in component.clusters})

        scores = {c: -s for c, s in scores.items()}

        return scores


__all__ = [
    "ClusterScores",
    "InstanceScores",
    "GraphScorer",
    "ClusterCardinality",
    "ComponentCardinality",
    "VertexDegree",
    "ParentCardinality",
    "GraphNeighborhood",
    "StationaryProbabilities",
]



"""Provides the core modules for CLAM."""

from . import cluster
from . import cluster_criteria
from . import dataset
from . import graph
from . import graph_criteria
from . import metric
from . import space
from .cluster import RATIO_NAMES
from .cluster import Cluster
from .cluster import Ratios
from .cluster_criteria import ClusterCriterion
from .dataset import Dataset
from .graph import Edge
from .graph import Graph
from .graph_criteria import GraphCriterion
from .metric import Metric
from .space import Space



"""This module defines the `Dataset` class."""

import abc
import pathlib
import random
import typing

import numpy


class Dataset(abc.ABC):
    """A `Dataset` is a collection of `instance`s.

    This abstract class provides utilities for accessing instances from the data.
    A `Dataset` should be combined with a `Metric` to produce a `Space`.
    """

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Ideally, a user would supply a unique name for each `Dataset`."""
        pass

    @property
    @abc.abstractmethod
    def data(self) -> typing.Any:  # noqa: ANN401
        """Returns the underlying data."""
        pass

    @property
    @abc.abstractmethod
    def indices(self) -> numpy.ndarray:
        """Returns the indices of the instances in the data."""
        pass

    @abc.abstractmethod
    def __eq__(self, other: "Dataset") -> bool:  # type: ignore[override]
        """Ideally, this would be a quick check based on `name`."""
        pass

    @property
    def cardinality(self) -> int:
        """The number of instances in the data."""
        return self.indices.shape[0]

    @property
    @abc.abstractmethod
    def max_instance_size(self) -> int:
        """The maximum memory size (in bytes) of any instance in the data."""
        pass

    @property
    @abc.abstractmethod
    def approx_memory_size(self) -> int:
        """Returns the approximate memory size (in bytes) of the data."""
        pass

    @abc.abstractmethod
    def __getitem__(
        self,
        item: typing.Union[int, typing.Iterable[int]],
    ) -> typing.Any:  # noqa: ANN401
        """Returns the instance(s) at the given index/indices."""
        pass

    @abc.abstractmethod
    def subset(
        self,
        indices: list[int],
        subset_name: str,
    ) -> "Dataset":
        """Returns a subset of the data using only the indexed instances.

        This subset should be of the same class as object from which it is created.
        """
        pass

    def max_batch_size(self, available_memory: int) -> int:
        """Conservatively estimate of the number of instances that will fill memory.

        Args:
            available_memory: in bytes.

        Returns:
            The number of instances that will fill `available_memory`.
        """
        return available_memory // self.max_instance_size

    def complement_indices(self, indices: list[int]) -> list[int]:
        """Returns the indices in the dataset that are not in the given `indices`."""
        return list(set(range(self.cardinality)) - set(indices))

    def subsample_indices(self, n: int) -> tuple[list[int], list[int]]:
        """Randomly subsample n indices from the dataset.

        Args:
            n: Number of indices to select.

        Returns:
            A 2-tuple of:
                - list of selected indices.
                - list of remaining indices.
        """
        if not (0 < n < self.cardinality):
            msg = (
                f"`n` must be a positive integer smaller than the cardinality of "
                f"the dataset ({self.cardinality}). Got {n} instead."
            )
            raise ValueError(msg)

        # TODO: This should be from self.indices
        indices = list(range(n))
        random.shuffle(indices)

        return indices[:n], indices[n:]


class TabularDataset(Dataset):
    """This wraps a 2d numpy array whose rows are instances and columns are features.

    To check if two `TabularDataset`s are equal, this class only checks that
    they have the same `name`.

    To check if two instances are equal, this class simply checks for
    element-wise equality among the instances.
    """

    def __init__(self, data: numpy.ndarray, name: str) -> None:
        """Initializes a `TabularDataset`.

        Args:
            data: A 2d array whose rows are instances and columns are features.
            name: This should be unique for each object the user instantiates.
        """
        self.__data = data
        self.__name = name
        self.__indices = numpy.asarray(list(range(data.shape[0])), dtype=numpy.uint)

    @property
    def name(self) -> str:
        """Returns the name of the dataset."""
        return self.__name

    @property
    def data(self) -> numpy.ndarray:
        """Returns the underlying data."""
        return self.__data

    @property
    def indices(self) -> numpy.ndarray:
        """Returns the indices of the instances in the data."""
        return self.__indices

    def __eq__(self, other: "TabularDataset") -> bool:  # type: ignore[override]
        """Checks if two `TabularDataset`s have the same name."""
        return self.__name == other.__name

    @property
    def max_instance_size(self) -> int:
        """Returns the maximum memory size (in bytes) of any instance in the data."""
        item_size = self.__data.itemsize
        num_items = numpy.prod(self.__data.shape[1:])
        return int(item_size * num_items)

    @property
    def approx_memory_size(self) -> int:
        """Returns the approximate memory size (in bytes) of the data."""
        return self.cardinality * self.max_instance_size

    def __getitem__(
        self,
        item: typing.Union[int, typing.Iterable[int]],
    ) -> numpy.ndarray:
        """Returns the instance(s) at the given index/indices."""
        return self.__data[item]

    def subset(self, indices: list[int], subset_name: str) -> "TabularDataset":
        """Returns a subset of the data using only the indexed instances."""
        return TabularDataset(self.__data[indices, :], subset_name)


class TabularMMap(TabularDataset):
    """This is identical to a `TabularDataset` but the data is a `numpy.memmap`."""

    def __init__(
        self,
        file_path: typing.Union[pathlib.Path, str],
        name: str,
        indices: typing.Optional[list[int]] = None,
    ) -> None:
        """Initializes a `TabularMMap`."""
        self.file_path = file_path
        data = numpy.load(str(file_path), mmap_mode="r")
        indices = list(range(data.shape[0])) if indices is None else indices
        self.__indices = numpy.asarray(indices, dtype=numpy.uint)
        if self.__indices.max(initial=0) >= data.shape[0]:
            msg = "Invalid indices provided."
            raise IndexError(msg)

        super().__init__(data, name)

    @property
    def indices(self) -> numpy.ndarray:
        """Returns the indices of the instances in the data."""
        return self.__indices

    def __getitem__(
        self,
        item: typing.Union[int, typing.Iterable[int]],
    ) -> numpy.ndarray:
        """Returns the instance(s) at the given index/indices."""
        indices = self.__indices[item]
        return numpy.asarray(self.data[indices, :])

    def subset(self, indices: list[int], subset_name: str) -> "TabularMMap":
        """Returns a subset of the data using only the indexed instances."""
        return TabularMMap(  # type: ignore[return-value]
            self.file_path,
            subset_name,
            indices,
        )


__all__ = [
    "Dataset",
    "TabularDataset",
    "TabularMMap",
]



"""Provides the Meta-ML models for CHAODA."""

import abc
import typing

import numpy
from sklearn import linear_model
from sklearn import tree

from ..core import cluster
from ..utils import helpers

logger = helpers.make_logger(__name__)


class MetaMLModel(abc.ABC):
    """A wrapper around a machine learning model."""

    def __init__(
        self,
        model_class: typing.Any,  # noqa: ANN401
        **kwargs,  # noqa: ANN003
    ) -> None:
        """Creates a model and initializes it with the given key-word arguments."""
        self.model = model_class(**kwargs)

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Returns the name of the model."""
        pass

    @abc.abstractmethod
    def fit(self, *args, **kwargs) -> "MetaMLModel":  # noqa: ANN002,ANN003
        """Fits the model."""
        pass

    @abc.abstractmethod
    def predict(self, ratios: numpy.ndarray) -> float:
        """Predicts using the underlying model."""
        pass

    @abc.abstractmethod
    def extract_python(
        self,
        *args,  # noqa: ANN002
        **kwargs,  # noqa: ANN003
    ) -> tuple[list[str], str]:
        """Extracts the scoring function as a string which can be written to disk.

        The scoring function should have the following signature:

        ```
        helpful_function_name(ratios: numpy.ndarray) -> float
        ```

        where `ratios` is a 1d array of floats with 6 elements. These ratios are
        `cluster_ratios` from Manifold.

        Returns:
            A 2-tuple whose items are:
            - a list of strings which will be written as the import lines.
            - a string containing the code for the scoring function in Python.
        """
        pass


class MetaDT(MetaMLModel):
    """A decision tree regressor with a maximum depth of 3."""

    def __init__(self) -> None:
        """Creates a decision tree regressor with a maximum depth of 3."""
        super().__init__(tree.DecisionTreeRegressor, max_depth=3)

    @property
    def name(self) -> str:
        """Returns the name of the model."""
        return "dt"

    def fit(self, data: numpy.ndarray, scores: numpy.ndarray) -> "MetaDT":
        """Fits the model."""
        logger.info(
            f"Fitting meta-model {self.name} on data with shape {data.shape} ...",
        )
        self.model = self.model.fit(data, scores)
        return self

    def predict(self, ratios: numpy.ndarray) -> float:
        """Predicts a score using the ratios of a Cluster."""
        return self.model.predict(ratios)

    def extract_python(self, metric: str, method: str) -> tuple[list[str], str]:
        """Extracts the scoring function as a string which can be written to disk.

        The string can be written to disk as a Python file.
        """
        # noinspection PyProtectedMember
        undefined_feature = tree._tree.TREE_UNDEFINED

        feature_names = [
            cluster.RATIO_NAMES[i] if i != undefined_feature else "undefined!"
            for i in self.model.tree_.feature
        ]

        # start function definition
        function_name = f"{self.name}_{metric}_{method}"
        code_lines: list[str] = [
            f"def {function_name}(ratios: numpy.ndarray) -> float:",
            f'    {", ".join(cluster.RATIO_NAMES)} = tuple(ratios)',
        ]

        def _extract_lines(node_index: int, indent: str) -> None:
            if (
                self.model.tree_.feature[node_index] != undefined_feature
            ):  # internal node
                feature: str = feature_names[node_index]
                threshold: float = self.model.tree_.threshold[node_index]
                left_index: int = self.model.tree_.children_left[node_index]
                right_index: int = self.model.tree_.children_right[node_index]

                # if block
                code_lines.append(f"{indent}if {feature} <= {threshold:.6e}:")
                _extract_lines(left_index, indent + "    ")

                # else block
                code_lines.append(f"{indent}else:")
                _extract_lines(right_index, indent + "    ")

            else:  # leaf node
                value: float = self.model.tree_.value[node_index][0][0]
                code_lines.append(f"{indent}return {value:.6e}")

        _extract_lines(0, "    ")
        code = "\n".join(code_lines)

        imports = ["import numpy"]

        return imports, code


class MetaLR(MetaMLModel):
    """A linear regression model without an intercept."""

    def __init__(self) -> None:
        """Creates a linear regression model without an intercept."""
        super().__init__(linear_model.LinearRegression, fit_intercept=False)

    @property
    def name(self) -> str:
        """Returns the name of the model."""
        return "lr"

    def fit(self, data: numpy.ndarray, scores: numpy.ndarray) -> "MetaLR":
        """Fits the model."""
        logger.info(
            f"Fitting meta-model {self.name} on data with shape {data.shape} ...",
        )
        self.model = self.model.fit(data, scores)
        return self

    def predict(self, ratios: numpy.ndarray) -> float:
        """Predicts a score using the ratios of a Cluster."""
        return self.model.predict(ratios)

    def extract_python(self, metric: str, method: str) -> tuple[list[str], str]:
        """Extracts the scoring function as a string which can be written to disk.

        The string can be written to disk as a Python file.
        """
        imports = ["import numpy"]

        function_name = f"{self.name}_{metric}_{method}"
        coefficients = [f"{float(c):.6e}" for c in self.model.coef_]
        code = "\n".join(
            [
                f"def {function_name}(ratios: numpy.ndarray) -> float:",
                "    return float(numpy.dot(numpy.asarray(",
                f'        a=[{", ".join(coefficients)}],',
                "        dtype=float,",
                "    ), ratios))",
            ],
        )

        return imports, code


__all__ = [
    "MetaMLModel",
    "MetaDT",
    "MetaLR",
]
